/* 
	Graph.java

	Title:			Kontsevich graph homology calculator
                                PhD thesis, Cornell University, Aug 2002
	Author:			Ferenc Gerlits
	Description:	this is the basic type for storing the graphs
*/

package Konts;

import java.util.Vector;

public class Graph extends Object implements java.io.Serializable
{
  public int size;			// number of vertices
  public byte m[][];		// to store the matrix 
  public int noEdges;		// # of edges (double edges are 1, loops 0)
  public int allEdges;		// # of all edges, with loops and all
  public int noLoops;		// # of loops; these always come first in NF
  public int noMultiples;	// # of vertices with multiple edges; these 
                            //  are next, before the rest of the vert's
  public int noTriangles;	// # of vertices contained in triangles (but
  							//	not in multiple edges)
  public int coeff;			// starts out 1, changes with permutation
  
  public Graph(int n) {
    size = n;
    m = new byte[size][size];
    noEdges = allEdges = noLoops = noMultiples = noTriangles = 0;
    coeff = 1;
    }
   
   
  public Graph(Graph g) {
    this(g.size);
    int i,j;
    
    for (i=0; i<size; i++)		// only storing #s on or above
      for (j=i; j<size; j++)	//   the diagonal
        m[i][j] = g.m[i][j];
    noEdges = g.noEdges;
    allEdges = g.allEdges;
    noLoops = g.noLoops;
    noMultiples = g.noMultiples;
    noTriangles = g.noTriangles;
    coeff = g.coeff;
    }
    

  public void permute(Permutation sigma) {
    int i, j, x, y, t;						// sigma is a permutation
    byte b[][] = new byte[size][size];		//   of 'size' numbers
    int inversions = 0;
    								
    for (i=0; i<size; i++)
      for (j=i; j<size; j++) {
        x = sigma.p[i];			
        y = sigma.p[j];			
        if (x > y) {			
          t = x;
          x = y;
          y = t;
          inversions += m[i][j];
          }
        b[x][y] = m[i][j];
        }	
    for (i=0; i<size; i++)
      for (j=i; j<size; j++)
        m[i][j] = b[i][j];
        
    if (inversions % 2 != 0)	// change of orientation of the edges...
      coeff = -coeff;
    coeff *= sigma.sign();		// ... and on the vertices
    }
    
    
  public boolean equals(Graph otherg) {
    int i, j;
    
    if (size != otherg.size ||
        noEdges != otherg.noEdges ||
        allEdges != otherg.allEdges ||
        noLoops != otherg.noLoops ||
        noMultiples != otherg.noMultiples ||
        noTriangles != otherg.noTriangles)
      return false;
    
    for (i=0; i<size; i++)
      for (j=i; j<size; j++)
        if (m[i][j] != otherg.m[i][j])
          return false;
    return true;
    }
    
    
  public static int order(Graph g1, Graph g2) {
    int i, j;			
    								// return 0 if the graphs are equal,
    for (i=0; i<g1.size; i++)		// -1 if g1<g2 and 1 if g1>g2
      for (j=i; j<g1.size; j++) {
        if (g1.m[i][j] < g2.m[i][j])	// not checking size; this 
          return -1;					// should only be called
        if (g1.m[i][j] > g2.m[i][j])	// when they're equal
          return 1;
        }
    return 0;
    }


  public Permutation normalize() {
    int i, j, k, n, nn, buf[];
    Permutation s;
        
    noEdges = 0;					// find correct value of noEdges,
    buf = new int[size];			//  noLoops, noMultiples and
    for (i=0; i<size; i++) 			//  noTriangles;
      if (m[i][i] != 0)				//  move special vertices to
        buf[i] = 1;					//  the front
    
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++) {
        if (m[i][j] != 0)
          noEdges++;
        if (m[i][j] > 1) {
          if (buf[i] == 0)
            buf[i] = 2;
          if (buf[j] == 0)
            buf[j] = 2;
          }
        }
    
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++)
        for (k=j+1; k<size; k++) {
          if (m[i][j] == 1 && m[i][k] == 1 && m[j][k] == 1) {
            if (buf[i] == 0)
              buf[i] = 3;
            if (buf[j] == 0)
              buf[j] = 3;
            if (buf[k] == 0)
              buf[k] = 3;
          }
        }
      
    s = new Permutation(size);
    n = 0;
    for (i=0; i<size; i++)
      if (buf[i] == 1)
        s.p[i] = n++;
    noLoops = nn = n;
    
    for (i=0; i<size; i++)
      if (buf[i] == 2)
        s.p[i] = n++;
    noMultiples = n - nn;
    nn = n;
    
    for (i=0; i<size; i++)
      if (buf[i] == 3)
        s.p[i] = n++;
    noTriangles = n - nn;
    
    for (i=0; i<size; i++)
      if (buf[i] == 0)
        s.p[i] = n++;
        
    permute(s);
    return s;
    }
    

  public Permutation toNormalForm(int stage) {
    int i, j, r;					// put graph in normal form
    Graph g, gmax;					// stage 1 = graph may have loops;
    								// stage 2 = loopless graphs
    Permutation sigma, sigma1, sigma2, sigma3, sigma4;
    Permutation sstored, tau;
    boolean oddSymm = false;
    
    tau = normalize();
    sigma1 = new Permutation(noLoops);
    sigma2 = new Permutation(noMultiples);
    sigma3 = new Permutation(noTriangles);
    sigma4 = new Permutation(size - noLoops - noMultiples - noTriangles);
    gmax = new Graph(this);
    sstored = new Permutation(size);

    do {
     do {
      do {
       do {
         sigma = Permutation.concatenate(sigma1, sigma2, sigma3, sigma4);
         g = new Graph(this);
         g.permute(sigma);
         if ((r = order(g, gmax)) == 1) {		// new lexico-
           sstored = sigma;						// graphic max
           gmax = g;				
           } 						
         else if (r == 0 && g.coeff != gmax.coeff) {	// found an odd
           if (stage == 1)								// automorphism
             oddSymm = true;
           else {
             coeff = 0;									
             return null;
             }
           }
        } while (sigma4.next());
       } while (sigma3.next());
      } while (sigma2.next());
     } while (sigma1.next());
    
    for (i=0; i<size; i++)		
      for (j=i; j<size; j++)
        m[i][j] = gmax.m[i][j];
    coeff = gmax.coeff;
    
    if (oddSymm)
      return null;
    else return Permutation.multiply(tau, sstored);
    }

  
  public int[] kthEdgeWithMult(int k) {	// find kth edge of the graph
    int i, j, n, 						//  with multiplicity
        r[] = new int[2];
        
    n = -1;
    for (i=0; i<size; i++)
      for (j=i; j<size; j++) {
        n += m[i][j];
        if (n >= k) {
          r[0] = i;
          r[1] = j;
          return r;
          }
        }
    r[0] = r[1] = -1;				// error: there is no kth edge
    return r;
    }


  public int[] kthEdgeWOMult(int k) {	// find kth edge of the graph
    int i, j, n, 						//  without multiplicity
        r[] = new int[2];
        
    n = -1;
    for (i=0; i<size; i++)
      for (j=i; j<size; j++) {	// including loops; is this correct?
        if (m[i][j] != 0)
          n++;
        if (n >= k) {
          r[0] = i;
          r[1] = j;
          return r;
          }
        }
    r[0] = r[1] = -1;				// error: there is no kth edge
    return r;
    }


  public void contractEdge(int k) {
    int r[];						// find the index of the
    								// kth edge, and call
    r = this.kthEdgeWOMult(k);			// the version below    
    contractEdge(r[0], r[1]);
    }
  

  public void contractEdge(int x, int y) {
    int i, j;
    
    if (x == y)	{		// can't contract a loop!
      System.out.println("Attempt to contract a loop in Graph.contractEdge.");
      System.exit(1);
      }
      
    byte b[][] = new byte[size - 1][size - 1];
        
    for (i=0; i<x; i++) {			// rows 1 through x-1
      for (j=i; j<x; j++)
        b[i][j] = m[i][j];
      b[i][x] = (byte)(m[i][x] + m[i][y]);
      for (j=x+1; j<y; j++)
        b[i][j] = m[i][j];
      for (j=y+1; j<size; j++)
        b[i][j-1] = m[i][j];
      }
    
    b[x][x] = (byte)(m[x][x] + m[y][y] + m[x][y] - 1);	// row x
    for (j=x+1; j<y; j++)
      b[x][j] = (byte)(m[x][j] + m[j][y]);
    for (j=y+1; j<size; j++)
      b[x][j-1] = (byte)(m[x][j] + m[y][j]);
      
    for (i=x+1; i<y; i++) {			// rows x+1 through y-1
      for (j=i; j<y; j++)
        b[i][j] = m[i][j];
      for  (j=y+1; j<size; j++)
        b[i][j-1] = m[i][j];
      }
      
    for (i=y+1; i<size; i++)		// rows y+1 through 'size'
      for (j=i; j<size; j++)
        b[i-1][j-1] = m[i][j];
        
    int inversions = 0;
    for (i=x+1; i<y; i++)			// sign change on the edges
      inversions += m[i][y];
    if (inversions % 2 != 0)
      coeff = -coeff;
    
    size--;
    m = b;    
    allEdges--;
    // g.noEdges, g.noLoops, etc. need to be fixed
    }  
    
  
  public Graph addEdge(int edge1, int edge2) {
    int i, j, r[];
    Graph g = new Graph(size + 2);	// add an extra edge between
									// the midpoints of the edge1th
    for (i=0; i<size; i++)			// and the edge2th edges
      for (j=i; j<size; j++)
        g.m[i][j] = m[i][j];
        
    if (edge1 != edge2) {    
      r = this.kthEdgeWithMult(edge1);
      g.m[r[0]][r[1]]--;	
      g.m[r[0]][size]++;	// this should work even when 
      g.m[r[1]][size]++;	//  edge1 is a loop
         
      r = this.kthEdgeWithMult(edge2);
      g.m[r[0]][r[1]]--;
      g.m[r[0]][size+1]++;
      g.m[r[1]][size+1]++;
    
      g.m[size][size+1] = 1;
      }
    else {
      r = this.kthEdgeWithMult(edge1);
      g.m[r[0]][r[1]]--;
      g.m[r[0]][size]++;
      g.m[r[1]][size+1]++;    
      g.m[size][size+1] = 2;
      }
      
    g.allEdges = allEdges + 3;	// -2 +5 or -1 +4  net change
    // g.noEdges, g.noLoops etc need to be fixed later

    return g;
    }


  public Graph addALoop(int edge) {
    int i, j, r[];
    Graph g = new Graph(size + 2);	// add an extra spoon (edge plus
									// a loop) at the midpoint of the
    for (i=0; i<size; i++)			// edge'th edge (w/o multiplicity)
      for (j=i; j<size; j++)
        g.m[i][j] = m[i][j];
        
    r = this.kthEdgeWOMult(edge);
    g.m[r[0]][r[1]]--;	
    g.m[r[0]][size]++;	// this should work even when 
    g.m[r[1]][size]++;	//  the edge is a loop
   
    g.m[size][size+1] = 1;
    g.m[size+1][size+1] = 1;
      
    g.allEdges = allEdges + 3;	// -1 +4  net change
    // g.noEdges, g.noLoops etc need to be fixed later

    return g;
    }


  public String toString() {
    int i, j;
    String str = new String();
    
    for (i=0; i<size; i++) {
      str = str + "[";
      for (j=0; j<i; j++)
        str = str + "\t";
      for (j=i; j<size; j++) { 
        str = str + "\t" + m[i][j];
        }
      str = str + "\t]\r\n";
      }
      
//  str = str + noEdges +", "+ allEdges +", "+ noLoops +", "+ noMultiples +", "+ noTriangles;
    return str;
    }  
}
